'''
Field module. Fields are grids with length scales.
'''

from ._base import Field
from ._complex import ComplexField
from ._real import RealField

from ._1d import ComplexField1D, RealField1D, Field1D

from ._2d import ComplexField2D, RealField2D, Field2D




